/********************************************************/
/*  This is the driver for the ALICE ALTRO USB core	*/
/*							*/
/*  Markus Joos, CERN-PH-ESS 15.3.2004			*/		
/*							*/
/*** C 2004 - The software with that certain something **/

/****************************************************************************************/
/* This dirver wors for kernels from 2.4 onwards					*/
/*											*/
/* Currently only one ALTRO card is supported. Extending the driver for multiple cards	*/
/* would require to introduce a "minor" array as done in the usb-skeleton driver	*/
/****************************************************************************************/

#include <linux/config.h>
#include <linux/module.h>
#include <linux/kernel.h>
#include <linux/slab.h>
#include <linux/init.h>
#include <linux/usb.h>
#include <linux/proc_fs.h>
#include <linux/devfs_fs_kernel.h>
#include <asm/uaccess.h>
#include "altro_common.h"

MODULE_PARM (debug, "i");
MODULE_PARM_DESC(debug, "1 = enable debugging   0 = disable debugging");
MODULE_PARM (timeout, "i");
MODULE_PARM_DESC(timeout, "timeout for usb_bulk_msg in units of jiffies");
MODULE_PARM (ddump, "i");
MODULE_PARM_DESC(ddump, "1 = enable data dumping   0 = disable data dumping");
MODULE_DESCRIPTION("USB driver for the ALICE ALTRO USB core");
MODULE_AUTHOR("Markus Joos, CERN/EP");
MODULE_LICENSE("Private: Contact markus.joos@cern.ch");


#ifndef USB_ALTRO_MINOR_BASE
# define USB_ALTRO_MINOR_BASE 192
#endif
#ifdef ALTRO_KERNEL_DEBUG
# define kdebug(x) {if (debug) printk x;}
#else
# define kdebug(x)
#endif

static int altro_open (struct inode *inode, struct file *file);
static int altro_release(struct inode *inode, struct file *file);
static int proc_write_altro(struct file *file, const char *buffer, unsigned long count, void *data);
static int proc_read_altro(char *page, char **start, off_t off, int count, int *eof, void *data);
static int altro_ioctl(struct inode *inode, struct file *file, unsigned int cmd, unsigned long arg);
static void altro_disconnect(struct usb_device *udev, void *clientdata);
static void *altro_probe(struct usb_device *udev, unsigned int ifnum, const struct usb_device_id *id);
int altro_init(void);
void altro_exit(void);

/* Globals */ 
static int cardtype, debug = 1, timeout = HZ * 10, ddump = 0;
static struct proc_dir_entry *altro_file;
struct altro_proc_data_t altro_proc_data;
struct altro_device *altrop = NULL;
int urbdone;

/* Externals */
extern devfs_handle_t usb_devfs_handle;

/*We need a local data structure, as it must be allocated for each*/
/*new device plugged in the USB bus*/ 
struct altro_device 
{
  char data [10];
  char *name;                          /* the name of the device */
  struct urb urb;                      /* USB Request block, to get USB data*/
  int maxp;                            /* packet len */
  char output[80];                     /* used for printk at irq time */
  unsigned char minor;                 /* the starting minor number for this device */
  devfs_handle_t devfs;                /* devfs device node */
  struct usb_device *udev;             /* save off the usb device pointer */
  struct semaphore sem;                /* locks this structure */
  unsigned char *bulk_in_buffer;       /* the buffer to receive data */
  int bulk_in_size;                    /* the size of the receive buffer */
  __u8 bulk_in_endpointAddr;           /* the address of the bulk in endpoint */
  unsigned char *bulk_out_buffer;      /* the buffer to send data */
  int bulk_out_size;                   /* the size of the send buffer */
  struct urb *write_urb;               /* the urb used to send data */
  __u8 bulk_out_endpointAddr;          /* the address of the bulk out endpoint */
};

/* Data in /proc file system */
struct altro_proc_data_t
{
  char name[10];
  char value[100];
};

/* The id_table, lists all devices that can be handled by this driver. */
/* The three numbers are class, subclass, protocol. <linux/usb.h> has */
/* more details about interface matches and vendor/device matches. */
/* Here we use a fake usb_device_id structure defined in ./usb-sysdep.h */

/*The parameters for the ALTRO are*/
/*Class    = */
/*Subclass = */
/*Protocol = */

static struct usb_device_id altro_id_table [] = 
{
  { 
    USB_DEVICE(0x1556, 0x0350),
    driver_info: (unsigned long)"altro_by_vendor (invalid device id)"
  },

  {
    USB_DEVICE(0x1556, 0x15e),
    driver_info: (unsigned long)"ALICE U2F"
  },

  { 
    USB_DEVICE(0x1556, 0x15f),
    driver_info: (unsigned long)"ALICE GOOFIE"
  },
  {
    USB_DEVICE(0x1556, 0x163),
    driver_info: (unsigned long)"ALICE MPT"
  },
  {
    0, /* no more matches */
  }
};

MODULE_DEVICE_TABLE (usb, altro_id_table);

/* file operations supported */
static struct file_operations altro_fops = 
{
  owner:       THIS_MODULE,
  ioctl:       altro_ioctl,
  open:        altro_open,
  release:     altro_release,
};

/* The callbacks are registered within the USB subsystem using the */
/* usb_driver data structure */
static struct usb_driver altro_usb_driver = 
{
  owner:       THIS_MODULE,
  name:        "altro",
  probe:       altro_probe,
  disconnect:  altro_disconnect,
  fops:        &altro_fops,
  minor:       USB_ALTRO_MINOR_BASE,
  id_table:    altro_id_table,
};


/************************************************************/
static int altro_open (struct inode *inode, struct file *file)
/************************************************************/
{
  int retval = 0;
  
  kdebug(("altro_drv(altro_open): function called\n"));
  if (altrop == NULL) 
  {
    kdebug(("altro_drv(altro_open): Sorry. So far no ALTRO has been connected\n"));
    return -ENODEV;
  }
  
  kdebug(("altro_drv(altro_open): minor number = %d\n", MINOR(inode->i_rdev)));
  
  /* save our object in the file's private structure */
  file->private_data = altrop;
  
  kdebug(("altro_drv(altro_open): function done\n"));
  return retval;
}


/**************************************************************/
static int altro_release(struct inode *inode, struct file *file)
/**************************************************************/
{
  kdebug(("altro_drv(altro_release): function called\n"));

  return(0);
}


/* These two callbacks are invoked when an USB device is detached or attached to the bus*/
/*********************************************************************/
static void altro_disconnect(struct usb_device *udev, void *clientdata)
/*********************************************************************/
{
  /* the clientdata is the altro_device we passed originally */
  struct altro_device *altro = clientdata;

  kdebug(("altro_drv(altro_disconnect): function called\n"));

  /* remove the URB, remove the input device, free memory */
  usb_unlink_urb(&altro->urb);

  /* remove our devfs node */
  devfs_unregister(altro->devfs);

  kfree(altro);
  kdebug(("altro_drv(altro_disconnect): USB %s disconnected\n", altro->name));

  return;
}


/*These two callbacks are invoked when an USB device is detached or attached to the bus */ 

/***************************************************************************************************/
static void *altro_probe(struct usb_device *udev, unsigned int ifnum, const struct usb_device_id *id)
/***************************************************************************************************/
{
  /* The probe procedure is pretty standard. Device matching has already */
  /* been performed based on the id_table structure */

  struct usb_interface *iface;
  struct usb_interface_descriptor *interface;
  struct usb_endpoint_descriptor *endpoint;
  struct altro_device *altro;
  int i, buffer_size, inok, outok;

  kdebug(("altro_drv(altro_probe): probe called for %s device\n", (char *)id->driver_info));
  kdebug(("altro_drv(altro_probe): ifnum = %d\n", ifnum));

  /* allocate and zero a new data structure for the new device */
  altro = kmalloc(sizeof(struct altro_device), GFP_KERNEL);
  if (!altro) 
  { 
    kdebug(("altro_drv(altro_probe): ERROR failed to kmalloc altro_device\n"));
    return NULL; /* failure */
  }
  
  memset(altro, 0, sizeof(*altro));
  altro->name = (char *)id->driver_info;
  init_MUTEX(&altro->sem);

  kdebug(("altro_drv(altro_probe): udev->descriptor.idVendor = 0x%08x\n", udev->descriptor.idVendor));
  kdebug(("altro_drv(altro_probe): udev->descriptor.idProduct = 0x%08x\n", udev->descriptor.idProduct));
  cardtype = udev->descriptor.idProduct;
  
  iface = &udev->actconfig->interface[ifnum];
  interface = &iface->altsetting[iface->act_altsetting];

  if (interface->bNumEndpoints != 2) 
  {
    kdebug(("altro_drv(altro_probe): ERROR interface->bNumEndpoints != 2 (number = %d)\n", interface->bNumEndpoints));
    return NULL;
  }
  
  inok = 0;
  outok = 0;
    
  for (i = 0; i < interface->bNumEndpoints; ++i) 
  {
    kdebug(("altro_drv(altro_probe): Processing endpoint %d ...\n", i));
    endpoint = &interface->endpoint[i];
    
    if ((endpoint->bEndpointAddress & 0x80) && ((endpoint->bmAttributes & 3) == 0x02)) 
    {
      if (!inok) 
      {
        kdebug(("altro_drv(altro_probe): we found the first bulk in endpoint\n"));
	buffer_size = endpoint->wMaxPacketSize;
	kdebug(("altro_drv(altro_probe): buffer_size = %d\n", buffer_size));
	altro->bulk_in_size = buffer_size;
	altro->bulk_in_endpointAddr = endpoint->bEndpointAddress;
	kdebug(("altro_drv(altro_probe): bulk_in_endpointAddr = %d\n", altro->bulk_in_endpointAddr));
	altro->bulk_in_buffer = kmalloc(buffer_size, GFP_KERNEL);
	if (!altro->bulk_in_buffer) 
	{
          kdebug(("altro_drv(altro_probe): Couldn't allocate bulk_in_buffer"));
          return NULL;
	}
	inok = 1;
      }
      else
        kdebug(("altro_drv(altro_probe): we found another bulk in endpoint\n"));
    }
    
    else if (((endpoint->bEndpointAddress & 0x80) == 0x00) && ((endpoint->bmAttributes & 3) == 0x02)) 
    {
      if (!outok) 
      {
	kdebug(("altro_drv(altro_probe): we found the first bulk out endpoint\n"));
	altro->write_urb = usb_alloc_urb(0);
	if (!altro->write_urb) 
	{
          kdebug(("altro_drv(altro_probe): No free urbs available"));
          return NULL;
	}
	buffer_size = endpoint->wMaxPacketSize;
	kdebug(("altro_drv(altro_probe): buffer_size = %d\n", buffer_size));
	altro->bulk_out_size = buffer_size;
	altro->bulk_out_endpointAddr = endpoint->bEndpointAddress;
	kdebug(("altro_drv(altro_probe): bulk_out_endpointAddr = %d\n", altro->bulk_out_endpointAddr));
	altro->bulk_out_buffer = kmalloc (buffer_size, GFP_KERNEL);
	if (!altro->bulk_out_buffer) 
	{
          kdebug(("altro_drv(altro_probe): Couldn't allocate bulk_out_buffer"));
          return NULL;
	}
	outok = 1;
      }	

      else
        kdebug(("altro_drv(altro_probe): we found another bulk out endpoint\n"));
    }
    else
      {
	kdebug(("altro_drv(altro_probe): we found an unidentifiable endpoint\n"));
	return NULL;
    }
  }
  
  altro->udev = udev;
  altrop = altro;
  kdebug(("altro_drv(altro_probe): End of function\n"));

  return altro;
}


/* Functions called at module load and unload time: only register and unregister the USB callbacks */
/******************/
int altro_init(void)
/******************/
{
  kdebug(("altro_drv(altro_init): function called\n"));

  /* Install /proc entry */
  altro_file = create_proc_entry("altro", 0644, NULL);
  if (altro_file == NULL) 
  {
    kdebug(("altro_drv(altro_init): ERROR from call to create_proc_entry\n"));
    return (-ENOMEM);
  }

  strcpy(altro_proc_data.name, "altro");
  strcpy(altro_proc_data.value, "altro");
  altro_file->data = &altro_proc_data;
  altro_file->read_proc = proc_read_altro;
  altro_file->write_proc = proc_write_altro;
  altro_file->owner = THIS_MODULE;
    
  return usb_register(&altro_usb_driver);
}

/* Functions called at module load and unload time: only register and */
/* unregister the USB callbacks */ 
/*******************/
void altro_exit(void)
/*******************/
{
  kdebug(("altro_drv(altro_exit): function called\n"));

  /* Remove /proc entry */
  remove_proc_entry("altro", NULL);
  
  usb_deregister(&altro_usb_driver);
}


/**********************/
module_init(altro_init);
/**********************/


/**********************/
module_exit(altro_exit);
/**********************/

/*************************************************************************************************/
static int proc_write_altro(struct file *file, const char *buffer, unsigned long count, void *data)
/*************************************************************************************************/
{
  int len;
  struct altro_proc_data_t *fb_data = (struct altro_proc_data_t *)data;

  kdebug(("altro_drv(proc_write_altro): proc_write_altro called\n"));
  
  if(count > 99)
    len = 99;
  else
    len = count;
  
  if (copy_from_user(fb_data->value, buffer, len)) 
  {
    kdebug(("altro_drv(proc_write_altro): error from copy_from_user\n"));
    return(-EFAULT);
  }

  kdebug(("altro_drv(proc_write_altro): len = %d\n", len));
  fb_data->value[len - 1] = '\0';
  kdebug(("altro_drv(proc_write_altro): text passed = %s\n", fb_data->value));
  
  if (!strcmp(fb_data->value, "debug")) 
  {
    debug = 1;
    kdebug(("altro_drv(proc_write_altro): debugging enabled\n")); 
  }
 
  if (!strcmp(fb_data->value, "nodebug")) 
  {
    kdebug(("altro_drv(proc_write_altro): debugging disabled\n")); 
    debug = 0; 
  }
  
  if (!strcmp(fb_data->value, "datadump")) 
  {
    ddump = 1;
    kdebug(("altro_drv(proc_write_altro): data dumping enabled\n")); 
  }
 
  if (!strcmp(fb_data->value, "nodatadump")) 
  {
    ddump = 0; 
    kdebug(("altro_drv(proc_write_altro): data dumping disabled\n")); 
  }
  
  if (!strcmp(fb_data->value, "dec")) 
  {
    MOD_DEC_USE_COUNT;
    kdebug(("altro_drv(proc_write_altro): Use count decremented\n"));
  }

  if (!strcmp(fb_data->value, "inc")) 
  {
    MOD_INC_USE_COUNT;
    kdebug(("altro_drv(proc_write_altro): Use count incremented\n"));
  }
 
  return len;
}


/**********************************************************************************************/
static int proc_read_altro(char *page, char **start, off_t off, int count, int *eof, void *data)
/**********************************************************************************************/
{
  int len;

  kdebug(("altro_drv(proc_read_altro): proc_read_altro called\n"));

  len = sprintf(page, "ALICE ALTRO USB driver\n");
  len += sprintf(page + len, " \n");
  len += sprintf(page + len, "Type of card connected: ");
  if (cardtype == 0x15e || cardtype == 0x350)
    len += sprintf(page + len, "ALICE U2F\n");
  else if (cardtype == 0x15f)
    len += sprintf(page + len, "ALICE GOOFIE\n");
  else if (cardtype == 0x163)
    len += sprintf(page + len, "ALICE MPT\n");
  else
    len += sprintf(page + len, "UNKNOWN (maybe your card is using an invalid device ID\n");
  len += sprintf(page + len, "\n");
  len += sprintf(page + len, "The time-out of usb_bulk_msg() is %d jiffies (usually: 1 jiffie = 10 ms)\n", timeout);
  len += sprintf(page + len, "\n");
  len += sprintf(page + len, "The command 'echo <action> > /proc/altro', executed as root,\n");
  len += sprintf(page + len, "allows you to interact with the driver. Possible actions are:\n");
  len += sprintf(page + len, "debug      -> enable debugging\n");
  len += sprintf(page + len, "nodebug    -> disable debugging\n");
  len += sprintf(page + len, "datadump   -> enable data dumping (use with care)\n");
  len += sprintf(page + len, "nodatadump -> disable data dumping\n");
  len += sprintf(page + len, "dec        -> Decrement use count\n");
  len += sprintf(page + len, "inc        -> Increment use count\n");

  *eof = 1;
  *start = page + off;
  len -= off;

  if (len > count)
    len = count;
  return (len);
}


/**************************************************************************************************/
static int altro_ioctl (struct inode *inode, struct file *file, unsigned int cmd, unsigned long arg)
/**************************************************************************************************/
{
  struct altro_device *altro;
  int pipe, retval, loop;
  
  kdebug(("altro_drv(altro_ioctl): function called\n")); 
  kdebug(("altro_drv(altro_ioctl): minor number = %d\n", MINOR (inode->i_rdev)));

  altro = (struct altro_device *)file->private_data;
  
  /* lock this object */
  down (&altro->sem);
  kdebug(("altro_drv(altro_ioctl): semaphore locked\n")); 
  
  /* verify that the device wasn't unplugged */
  if (altro->udev == NULL) 
  {
    up (&altro->sem);
    kdebug(("altro_drv(altro_ioctl): ERROR device was unplugged\n"));
    return -ENODEV;
  }
  kdebug(("altro_drv(altro_ioctl): The device is still connected\n")); 
  
  switch (cmd) 
  {    
    case GET_DATA: 
    {   
      altro_bulk_in_t params; /*MJ: this memory gets allocated on the stack. Better move it to the private data*/
      kdebug(("altro_drv(ioctl:GET_DATA): function called for endpoint 0x%08x\n", altro->bulk_in_endpointAddr));
      kdebug(("altro_drv(ioctl:GET_DATA): altro->bulk_in_size = %d\n", altro->bulk_in_size));

      pipe = usb_rcvbulkpipe(altro->udev, altro->bulk_in_endpointAddr);
      kdebug(("altro_drv(ioctl:GET_DATA): reading from pipe %d\n", pipe));

      /* do an immediate bulk read to get data from the device */
      /* MJ: The parameter altro->bulk_in_size requests as much data as the ALTRO can actually send */
      /* MJ: The real abount of data may be less than that.*/
      /* MJ: The constant BULK_IN_SIZE in the header file has to be tuned to the value of endpoint->wMaxPacketSize*/ 
      retval = usb_bulk_msg(altro->udev, pipe, &params.data[0], altro->bulk_in_size, &params.nbytes, timeout);
      kdebug(("altro_drv(ioctl:GET_DATA): usb_bulk_msg returns %d\n", retval));

      if (retval)  
      {
	kdebug(("altro_drv(ioctl:GET_DATA): No data received\n"));
	params.nbytes = 0;
      }
      else
	kdebug(("altro_drv(ioctl:GET_DATA): %d bytes received\n", params.nbytes));

      if (ddump)
      {
	kdebug(("altro_drv(ioctl:GET_DATA): Now dumping newly received data.......\n"));
        for (loop = 0; loop < params.nbytes; loop++)
	  kdebug(("altro_drv(ioctl:GET_DATA): Byte #%d = 0x%02x\n", loop, params.data[loop]));
      } 

      if (copy_to_user ((void *)arg, &params, sizeof(altro_bulk_in_t)) != 0) 
      {
	kdebug(("altro_drv(ioctl:GET_DATA): error from copy_to_user\n"));
	up (&altro->sem);
	return(-ALTRO_EFAULT);
      }
      break;
    }    

    case SEND_DATA: 
    {   
      altro_bulk_out_t params;  /*MJ: this memory gets allocated on the stack. Better move it to the private data*/
      int maxp, ndone;

      kdebug(("altro_drv(ioctl:SEND_DATA): function called for endpoint 0x%08x\n", altro->bulk_out_endpointAddr));
      kdebug(("altro_drv(ioctl:SEND_DATA): altro->bulk_in_size = %d\n", altro->bulk_out_size));

      if (copy_from_user(&params, (void *)arg, sizeof(altro_bulk_out_t)) !=0) 
      {
	kdebug(("altro_drv(ioctl:SEND_DATA): error from copy_from_user\n"));
	up (&altro->sem);
	return(-ALTRO_EFAULT);
      }

      pipe = usb_sndbulkpipe(altro->udev, altro->bulk_out_endpointAddr);
      kdebug(("altro_drv(ioctl:SEND_DATA): reading from pipe %d\n", pipe));

      maxp = usb_maxpacket(altro->udev, pipe, usb_pipeout(pipe));
      if (maxp > altro->bulk_out_size) 
	maxp = altro->bulk_out_size; 
      kdebug(("altro_drv(ioctl:SEND_DATA): it is possible to send %d bytes\n", maxp));

      /* check if we would write more data than possible */
      if (params.nbytes > maxp) 
      {	
	kdebug(("altro_drv(ioctl:SEND_DATA): Cannot write %d bytes to pipe of %d bytes\n", params.nbytes, maxp));
	up (&altro->sem);
	return(-ALTRO_EFAULT);
      }
      kdebug(("altro_drv(ioctl:SEND_DATA): trying to send %d bytes\n", params.nbytes));

      if (ddump)
      {
	kdebug(("altro_drv(ioctl:SEND_DATA): Dumping data berfore sending it.......\n"));
        for (loop = 0; loop < params.nbytes; loop++)
	  kdebug(("altro_drv(ioctl:SEND_DATA): Byte #%d = 0x%02x\n", loop, params.data[loop]));
      } 

      /* do an immediate bulk write to send data to the device */
      retval = usb_bulk_msg(altro->udev, pipe, &params.data[0], params.nbytes, &ndone, timeout);
      kdebug(("altro_drv(ioctl:SEND_DATA): usb_bulk_msg returns %d\n", retval));
      if (retval) 
      {
	kdebug(("altro_drv(ioctl:SEND_DATA): No data sent\n"));
	up (&altro->sem);
	return(-ALTRO_EFAULT);
      }
      else
	kdebug(("altro_drv(ioctl:SEND_DATA): %d bytes sent\n", ndone));

      break;
    }     

    case CONTROL: 
    {
      altro_control_t params;  /*MJ: this memory gets allocated on the stack. Better move it to the private data*/

      kdebug(("altro_drv(ioctl:CONTROL): function called\n"));

      if (copy_from_user(&params, (void *)arg, sizeof(altro_control_t)) !=0 ) 
      {
	kdebug(("altro_drv(ioctl:CONTROL): error from copy_from_user\n"));
	up (&altro->sem);
	return(-ALTRO_EFAULT);
      }

      kdebug(("altro_drv(ioctl:CONTROL): params.bRequest      = %d\n", params.bRequest));
      kdebug(("altro_drv(ioctl:CONTROL): params.bmRequestType = %d\n", params.bmRequestType));
      kdebug(("altro_drv(ioctl:CONTROL): params.wValue        = %d\n", params.wValue));
      kdebug(("altro_drv(ioctl:CONTROL): params.wIndex        = %d\n", params.wIndex));
      kdebug(("altro_drv(ioctl:CONTROL): params.data[0]       = %d\n", params.data[0]));
      kdebug(("altro_drv(ioctl:CONTROL): params.wLength       = %d\n", params.wLength));

      retval = usb_control_msg(altro->udev, usb_rcvctrlpipe(altro->udev,0), params.bRequest, params.bmRequestType, 
			       params.wValue, params.wIndex, &params.data[0], params.wLength, HZ);
      if (retval) 
      {
	kdebug(("altro_drv(ioctl:CONTROL): ERROR  failed to register control URB. Error code = %d\n", retval));
	up (&altro->sem); 
	return(-ALTRO_EFAULT);
      }
      break;
    }
  }
  /* unlock the device */
  up (&altro->sem);
  return(0);
}
